// Copyright (C) 1998 Bruce Guenter <bruce.guenter@qcc.sk.ca>
//
// This package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

#ifdef TRACE
#define trace(X) cerr << (void*)this << "->" << __PRETTY_FUNCTION__ << X << endl
#define trace_static(X) cerr << __PRETTY_FUNCTION__ << X << endl
#else
#define trace(X) do { } while(0)
#define trace_static(X) do { } while(0)
#endif

#include "mystring.h"
#include <ctype.h>

static mystringrep nil = { 0, 1, "" };

static const unsigned replength = sizeof(unsigned)*2;

mystringrep* mystringrep::alloc(unsigned length)
{
  trace_static("length=" << length);
  if(length == 0)
    return &nil;
  unsigned size = length + replength+1;
  mystringrep* ptr = (mystringrep*)new char[size];
  ptr->length = length;
  ptr->references = 1;
  return ptr;
}

mystringrep* mystringrep::dup(const char* str, unsigned length)
{
  trace_static("str=" << (void*)str << " length=" << length);
  if(length == 0)
    return &nil;
  mystringrep* ptr = alloc(length);
  memcpy(ptr->buf, str, length);
  ptr->buf[length] = 0;
  return ptr;
}

mystringrep* mystringrep::dup(const char* str1, unsigned length1,
			      const char* str2, unsigned length2)
{
  trace_static("");
  if(length1+length2 == 0)
    return &nil;
  mystringrep* ptr = alloc(length1+length2);
  memcpy(ptr->buf, str1, length1);
  memcpy(ptr->buf+length1, str2, length2);
  ptr->buf[length1+length2] = 0;
  return ptr;
}

void mystringrep::attach()
{
  trace("references=" << references);
  ++references;
}

void mystringrep::detach()
{
  trace("references=" << references);
  
  --references;
  if(!references) {
    trace("deleting this");
    delete this;
  }
}

///////////////////////////////////////////////////////////////////////////////

mystring::~mystring()
{
  trace("rep=" << (void*)rep);
  rep->detach();
}

void mystring::dup(mystringrep* in)
{
  trace("in=" << (void*)in);
  rep = in;
  rep->attach();
}

void mystring::dupnil()
{
  trace("");
  dup(&nil);
}

void mystring::dup(const char* in, size_t len)
{
  trace("len=" << len);
  if(!in || !len || !in[0])
    dupnil();
  else
    rep = mystringrep::dup(in, len);
}

void mystring::dup(const char* in)
{
  trace("in='" << in << "'");
  if(!in || !in[0])
    dupnil();
  else
    dup(in, strlen(in));
}
  
void mystring::dup(const char* x, const char* y)
{
  trace("");
  if(!x || !x[0]) dup(y);
  else if(!y || !y[0]) dup(x);
  else
    rep = mystringrep::dup(x, strlen(x), y, strlen(y));
}

void mystring::dup(const char* a, const char* b, const char* c)
{
  trace("");
  if(!a || !a[0]) dup(b,c);
  else if(!b || !b[0]) dup(a,c);
  else if(!c || !c[0]) dup(a,b);
  else {
    dup(a, b);
    append(c, strlen(c));
  }
}

#if 0  
void mystring::dup(const char* a, const char* b, const char* c, const char* d)
{
  if(!a || !a[0]) dup(b, c, d);
  else if(!b || !b[0]) dup(a, c, d);
  else if(!c || !c[0]) dup(a, b, d);
  else if(!d || !d[0]) dup(a, b, c);
  else {
    size_t la = strlen(a);
    size_t lb = strlen(b);
    size_t lc = strlen(c);
    size_t ld = strlen(d);
    alloc(la+lb+lc+ld);
    memcpy(buf, a, la);
    memcpy(buf+la, b, lb);
    memcpy(buf+la+lb, c, lc);
    memcpy(buf+la+lb+lc, d, ld);
    buf[la+lb+lc+ld] = 0;
    tracestr("mystring::dup/4", buf);
  }
}

void mystring::dup(const char* a, const char* b, const char* c,
		   const char* d, const char* e)
{
  if(!a || !a[0]) dup(b, c, d, e);
  else if(!b || !b[0]) dup(a, c, d, e);
  else if(!c || !c[0]) dup(a, b, d, e);
  else if(!d || !d[0]) dup(a, b, c, e);
  else if(!e || !e[0]) dup(a, b, c, d);
  else {
    size_t la = strlen(a);
    size_t lb = strlen(b);
    size_t lc = strlen(c);
    size_t ld = strlen(d);
    size_t le = strlen(e);
    alloc(la+lb+lc+ld+le);
    memcpy(buf, a, la);
    memcpy(buf+la, b, lb);
    memcpy(buf+la+lb, c, lc);
    memcpy(buf+la+lb+lc, d, ld);
    memcpy(buf+la+lb+lc+ld, e, le);
    buf[la+lb+lc+ld+le] = 0;
    tracestr("mystring::dup/5", buf);
  }
}
#endif

mystring mystring::operator=(const char* in)
{
  trace("in='" << in << "'");
  if(in != rep->buf) {
    rep->detach();
    dup(in);
  }
  return *this;
}

mystring mystring::operator=(const mystring& in)
{
  trace("in.rep=" << (void*)in.rep);
  if(in.rep != rep) {
    mystringrep* oldrep = rep;
    rep = in.rep;
    rep->attach();
    oldrep->detach();
  }
  return *this;
}

mystring mystring::subst(char from, char to) const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = (*in == from) ? to : *in;
  return mystring(buf, length);
}

mystring mystring::lower() const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = isupper(*in) ? tolower(*in) : *in;
  return mystring(buf, length);
}

mystring mystring::upper() const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = islower(*in) ? toupper(*in) : *in;
  return mystring(buf, length);
}

int mystring::find(char ch, size_t offset) const
{
  if(offset >= rep->length)
    return -1;
  char* ptr = strchr(rep->buf+offset, ch);
  return ptr ? ptr-rep->buf : -1;
}

int mystring::find_last(char ch) const
{
  char* ptr = strrchr(rep->buf, ch);
  return ptr ? ptr-rep->buf : -1;
}
  
int mystring::find_first(const char* set, size_t offset) const
{
  for(; offset < rep->length; offset++) {
    if(strchr(set, rep->buf[offset]))
      return offset;
  }
  return -1;
}

int mystring::find_last(const char* set) const
{
  for(int i = rep->length-1; i >= 0; --i) {
    if(strchr(set, rep->buf[i]))
      return i;
  }
  return -1;
}

// return the sub-string ending at 'offset'
mystring mystring::left(size_t offset) const
{
  if(offset > rep->length)
    return mystring();
  else
    return mystring(rep->buf, offset);
}

// return the sub-string starting at 'offset'
mystring mystring::right(size_t offset) const
{
  if(offset >= rep->length)
    return mystring();
  else
    return mystring(rep->buf+offset, rep->length-offset);
}

// return the 'len' characters of the string starting at 'offset'
mystring mystring::sub(size_t offset, size_t len) const
{
  // return right(offset).left(len);
  if(len == 0)
    return mystring();
  else {
    if(len+offset >= rep->length)
      len = rep->length - offset;
    return mystring(rep->buf+offset, len);
  }
}

void mystring::append(const char* str, size_t len)
{
  trace("");
  if(str && len > 0 && str[0]) {
    mystringrep* newrep = mystringrep::dup(rep->buf, rep->length, str, len);
    rep->detach();
    rep = newrep;
  }
}

mystring mystring::lstrip() const
{
  const char* ptr = rep->buf;
  while(*ptr && isspace(*ptr))
    ++ptr;
  return ptr;
}

mystring mystring::rstrip() const
{
  const char* ptr = rep->buf + rep->length - 1;
  while(ptr >= rep->buf && isspace(*ptr))
    --ptr;
  return mystring(rep->buf, ptr-rep->buf+1);
}

mystring mystring::strip() const
{
  const char* start = rep->buf;
  while(*start && isspace(*start))
    ++start;
  const char* end = rep->buf + rep->length - 1;
  while(end >= start && isspace(*end))
    --end;
  return mystring(start, end-start+1);
}
    
mystring operator+(const mystring& s1, const mystring& s2)
{
  mystring res = s1;
  res += s2;
  return res;
}

mystring operator+(const mystring& s1, const char* s2)
{
  mystring res = s1;
  res += s2;
  return res;
}

istream& operator>>(istream& in, mystring& str)
{
  str = "";
  char buf[256];		// buffer this many characters at a time
  unsigned i = 0;
  in >> ws;			// skip leading whitespace
  char ch;
  while((ch = in.get()) != EOF) {
    if(isspace(ch)) {		// end the input on whitespace
      buf[i] = 0;
      str += buf;
      i = 0;
      break;
    }
    else {
      buf[i++] = ch;
      if(i == sizeof(buf)-1) {	// append the filled buffer and empty it
	buf[i] = 0;
	str += buf;
	i = 0;
      }
    }
  }
  if(i > 0) {			// If EOF was reached before whitespace,
    buf[i] = 0;			// append the buffer to the string.
    str += buf;
  } else if(str.length() == 0)	// Mark failure if no string was read.
    in.set(ios::failbit);
  return in;
}
